/*
 * -- Copyright 2000 --
 * George Greer <greerga@m-l.org>
 * Corey Sanders <corey.sanders@email.com>
 *
 * Licensed under the GNU General Public License (GPL) v2 (only).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

#include "algol60.h"

static stack_t *ref_params;
static queue_t *decl_params;

void refparaminit(void)
{
  ref_params = newstack();
}

void refparamclean(void)
{
  deletestack(ref_params);
}

struct symtab *asm_procedure_decl(enum e_type type, char *procname)
{
  char *param_name;
  struct symtab *func = syminsert(root_block, procname, BYVAL, PROCEDURE);

  param_name = proc_param_name(procname, 0); // cleanup handled elsewhere
  syminsert(root_block, param_name, BYVAL, type);

  fprintf(outfile, ":%s\n", procname);

  assert(decl_params == NULL);
  decl_params = newqueue();
  return func;
}

void asm_procedure_decl_cleanup(char *procname)
{
  struct symtab *itr, *p;
  char *param_name;
  int paramnum;

  assert(decl_params);
  for (paramnum = 0; ; paramnum++) {
    param_name = proc_param_name(procname, paramnum);
    itr = symlook(root_block, param_name);
    free(param_name);
    if (!itr)
      break;
    if (itr->callby != BYREF)
      continue;
    p = dequeue(decl_params);
    fprintf(stderr, "%d: %s\n", paramnum, p->name);
    asm_assign(itr, p);
  }

  deletequeue(decl_params);
  decl_params = NULL;

  fputs("return\n", outfile);
}

struct symtab *asm_procedure_call(struct symtab *proc)
{
  struct symtab *retval;
  char *param_name;

  fprintf(outfile, "call :%s\n", proc->name);

  param_name = proc_param_name(proc->name, 0);

  retval = symlook(root_block, param_name);
  if( !retval ) { fprintf(stderr, "lookup failed for [%s]\n", param_name); symdump(root_block); }

  free(param_name);

  proc_param_ref_store(proc->name);

  return retval;
}

void asm_proc_call_param_check(struct symtab *proc, struct symtab *actual)
{
  static int param_num=1;
  static char *proc_name;

  struct symtab *formal;
  char *param_name;

  if(proc && !actual) {
    proc_name = proc->name;
    param_num = 1;
    return;
  }
  
  param_name = proc_param_name(proc_name, param_num);
  formal = symlook(root_block, param_name);
  free(param_name);

  if( !formal && actual ) {
    yyerror("too many arguments to function '%s'", unamp(proc_name));
  }

  if( !proc && !actual ) {
    if( formal != NULL )
      yyerror("too few parameters being passed to '%s'", unamp(proc_name));
  } else if( !proc && actual ) {
    if( formal->type != actual->type )
      yyerror("param %d of '%s' of type '%s' being passed as '%s'", param_num, unamp(proc_name), symtype(formal->type), symtype(actual->type));

    fprintf(outfile, "store %s,%s\n", actual->name, formal->name); // success
  }

  if( formal && formal->callby == BYREF )
    push(ref_params, actual);

  param_num++;
}

void asm_proc_decl_param(enum e_type type, char *formal, enum e_calltype callby, char *proc)
{
  static char *proc_name;
  static int proc_num = 1;
  struct symtab *refparm;
  char *internal;

  if( proc ) {
    proc_name = proc;
    proc_num = 1;
    return;
  }

  refparm = syminsert(root_block, formal, callby, type);
  fprintf(outfile, "declare %s,%s\n", formal, symtype(type));
  if (callby == BYREF)
    enqueue(decl_params, refparm);

  internal = proc_param_name(proc_name, proc_num++);
  syminsert(root_block, internal, callby, type);

  fprintf(outfile, "store %s,%s\n", internal, formal);
}

char *proc_param_name(const char *procname, int param_num)
{
  char *param_name;
  param_name = calloc(1, strlen(procname) + 10 + 2);

  if( ! param_name ) {
    fprintf(stderr, "out of memory error\n");
    exit(EXIT_FAILURE);
  }

  sprintf(param_name, "%s_%d", procname, param_num); // FUDGE!!!

  return param_name;
}

/* KLUGE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
** -- use a stack to reverse the order of the parameters coming off the original stack 
** -- a queue would not work because of nested procedure calls
*/
void proc_param_ref_store(const char *procname)
{
  char *paramname;
  int paramnum;
  struct symtab *formal, *actual;

  stack_t *refs;

  refs = newstack(); 

  paramnum = 0;
  do {
    paramname = proc_param_name(procname, paramnum);
    formal = symlook(root_block, paramname);
    
    if( formal && formal->callby == BYREF ) {
      actual = pop(ref_params);
      push(refs, actual);
    }

    free(paramname); // cleanup memory from proc_param_name

    paramnum++;
  } while( formal != NULL );

  paramnum = 0;
  do {
    paramname = proc_param_name(procname, paramnum);
    formal = symlook(root_block, paramname);

    if( formal && formal->callby == BYREF ) {
      actual = pop(refs);
      fprintf(outfile, "store %s,%s\n", formal->name, actual->name); // store from, to
    }

    free(paramname); // cleanup memory from proc_param_name

    paramnum++;
  } while( formal != NULL );

  deletestack(refs);
}

