/*
 * -- Copyright 2000 --
 * George Greer <greerga@m-l.org>
 * Corey Sanders <corey.sanders@email.com>
 *
 * Licensed under the GNU General Public License (GPL) v2 (only).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

#include "algol60.h"

int NSYMS = 128;

void syminit(struct block_s *block)
{
	static unsigned int block_level=0;

	block->symbol_table = (struct symtab *) calloc(1, sizeof(struct symtab) * NSYMS);
	if( block->symbol_table == NULL )
	{
		fprintf(stderr, "syminit: out of memory.\n");
		exit(EXIT_FAILURE);
	}
	block->level = block_level++;
}

void symclean(struct block_s *block)
{
	struct symtab *sp;
	for(sp=block->symbol_table; sp < &block->symbol_table[NSYMS]; sp++)
	{
		if( sp->name ) free(sp->name);
	}
}

struct symtab * symlook(struct block_s *block, const char *s)
{
	struct symtab *sp;
	int i, found;
	char *pos; //, *symname;

	assert(s != NULL);

	found = 0;
	for(i=0; block->symbol_table[i].name != NULL; i++)
	{
		sp = &block->symbol_table[i];

		pos = strrchr(sp->name, '&');
		if( !pos ) {
		  if( !strcmp(sp->name, s) ) // literal
		    return sp; //found = 1;
		} else if( (pos - sp->name) == strlen(s) && !strncmp(sp->name, s, strlen(s))) // variable
		  return sp; //found = 1;
		else if( strchr(pos, '_') && !strcmp(sp->name, s) ) // function parameter
		  return sp; //found = 1;

		//if( found ) return sp;
	}

	return (block != root_block) ? symlook(block->prev, s) : NULL;
}

struct symtab * syminsert(struct block_s *block, const char *name, enum e_calltype callby, enum e_type type)
{
	struct symtab *sp;

	for(sp=block->symbol_table; sp < &block->symbol_table[NSYMS]; sp++)
	{
		if(sp->name && !strcmp(sp->name, name)) {
			fprintf(stderr, "Attempt to duplicate symbol entry.");
			return NULL; /* error exit */
		}
		if(!sp->name)
		{
			//sp->name = strdup(name);
			sp->name = (char *)name; /* Lexer handles malloc */
			sp->type = type;
			sp->callby = callby;
			return sp; /* success */
		}
		/* otherwise, symbol table overflow */
	}

	/* on overflow, double the size of the table */
	NSYMS *= 2;	
	block->symbol_table = (struct symtab * ) realloc(block->symbol_table, sizeof(struct symtab) * NSYMS);
	if( block->symbol_table == NULL ) {
	    fprintf(stderr, "syminsert: out of memory\n");
	    exit(EXIT_FAILURE); /* symtab out of memory */
	} else {
	    memset(block->symbol_table + (NSYMS/2), 0, sizeof(struct symtab) * (NSYMS/2)); 
	    return syminsert(block, name, callby, type);
	}

	assert(0);
}

struct symtab *symtempvar(enum e_type type)
{
  static unsigned int temp_var = 0;
  char varname[16];	/* 2 unused */

  sprintf(varname, "T&&%u", temp_var++);
  fprintf(outfile, "declare %s,%s\n", varname, symtype(type));
  return syminsert(current_block, strdup(varname), BYVAL, type);
}

const char * symtype(enum e_type type)
{
  switch(type) {
    case INTEGER: return "integer"; break;
    case REAL: return "real"; break;
    case BOOL: return "bool"; break;
    case PROCEDURE : return "procedure" ; break;
    case VOID : return "void" ; break;
    default:
    	fprintf(stderr, "symtype: invalid symbol type");
    	exit(EXIT_FAILURE);
  }

  assert(0);
}

struct symtab * symiliteral(long val)
{
  char literal[16];
  struct symtab *symvar;

  sprintf(literal, "%ld", val);
  
  symvar = symlook(root_block, literal);
  if(symvar == NULL) 
  	return syminsert(root_block, strdup(literal), BYVAL, INTEGER);
  else
  	return symvar;
}

struct symtab * symrliteral(float val)
{
  char literal[16];
  struct symtab *symvar;

  sprintf(literal, "%f", val);
  symvar = symlook(root_block, literal);

  if(symvar == NULL) 
  	return syminsert(root_block, strdup(literal), BYVAL, REAL);
  else
  	return symvar;
}

int symdeclok(struct block_s *block, struct symtab *sp)
{
  char *pos = strrchr(sp->name, '&');
  if( pos != NULL )
    if(block->level == atoi(pos+1)) return 0;

  return 1;
}

void symdump(struct block_s *block)
{
  int i;

  fprintf(stderr, "-- symdump (block %d) --\n", block->level);
  for(i=0; block->symbol_table[i].name != NULL; i++) {
    fprintf(stderr, "%s: %s\n", block->symbol_table[i].name, symtype(block->symbol_table[i].type));
  }
}

void symdeclparam(void)
{
  struct symtab *sp;
  int i;
  char *p_amp, *p_und;

  for(i=0; root_block->symbol_table[i].name != NULL; i++)
  {
    sp = &root_block->symbol_table[i];

    p_amp = strrchr(sp->name, '&');
    if( p_amp ) {
      p_und = strchr(p_amp, '_');
    }
    if( p_amp && p_und && sp->type != VOID )
      fprintf(outfile, "declare %s,%s\n", sp->name, symtype(sp->type));
  }
}
